{
  "cells": [
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "%matplotlib inline"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n# Example: Switching Linear Dynamical System EEG\n\nWe use a switching linear dynamical system [1] to model a EEG time series dataset.\nFor inference we use a moment-matching approximation enabled by\n`funsor.interpretations.moment_matching`.\n\nReferences\n\n[1] Anderson, B., and J. Moore. \"Optimal filtering. Prentice-Hall, Englewood Cliffs.\" New Jersey (1979).\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import argparse\nimport time\nfrom collections import OrderedDict\nfrom os.path import exists\nfrom urllib.request import urlopen\n\nimport numpy as np\nimport pyro\nimport torch\nimport torch.nn as nn\n\nimport funsor\nimport funsor.ops as ops\nimport funsor.torch.distributions as dist\nfrom funsor.pyro.convert import (\n    funsor_to_cat_and_mvn,\n    funsor_to_mvn,\n    matrix_and_mvn_to_funsor,\n    mvn_to_funsor,\n)\n\n\n# download dataset from UCI archive\ndef download_data():\n    if not exists(\"eeg.dat\"):\n        url = \"http://archive.ics.uci.edu/ml/machine-learning-databases/00264/EEG%20Eye%20State.arff\"\n        with open(\"eeg.dat\", \"wb\") as f:\n            f.write(urlopen(url).read())\n\n\nclass SLDS(nn.Module):\n    def __init__(\n        self,\n        num_components,  # the number of switching states K\n        hidden_dim,  # the dimension of the continuous latent space\n        obs_dim,  # the dimension of the continuous outputs\n        fine_transition_matrix=True,  # controls whether the transition matrix depends on s_t\n        fine_transition_noise=False,  # controls whether the transition noise depends on s_t\n        fine_observation_matrix=False,  # controls whether the observation matrix depends on s_t\n        fine_observation_noise=False,  # controls whether the observation noise depends on s_t\n        moment_matching_lag=1,\n    ):  # controls the expense of the moment matching approximation\n\n        self.num_components = num_components\n        self.hidden_dim = hidden_dim\n        self.obs_dim = obs_dim\n        self.moment_matching_lag = moment_matching_lag\n        self.fine_transition_noise = fine_transition_noise\n        self.fine_observation_matrix = fine_observation_matrix\n        self.fine_observation_noise = fine_observation_noise\n        self.fine_transition_matrix = fine_transition_matrix\n\n        assert moment_matching_lag > 0\n        assert (\n            fine_transition_noise\n            or fine_observation_matrix\n            or fine_observation_noise\n            or fine_transition_matrix\n        ), (\n            \"The continuous dynamics need to be coupled to the discrete dynamics in at least one way [use at \"\n            + \"least one of the arguments --ftn --ftm --fon --fom]\"\n        )\n\n        super(SLDS, self).__init__()\n\n        # initialize the various parameters of the model\n        self.transition_logits = nn.Parameter(\n            0.1 * torch.randn(num_components, num_components)\n        )\n        if fine_transition_matrix:\n            transition_matrix = torch.eye(hidden_dim) + 0.05 * torch.randn(\n                num_components, hidden_dim, hidden_dim\n            )\n        else:\n            transition_matrix = torch.eye(hidden_dim) + 0.05 * torch.randn(\n                hidden_dim, hidden_dim\n            )\n        self.transition_matrix = nn.Parameter(transition_matrix)\n        if fine_transition_noise:\n            self.log_transition_noise = nn.Parameter(\n                0.1 * torch.randn(num_components, hidden_dim)\n            )\n        else:\n            self.log_transition_noise = nn.Parameter(0.1 * torch.randn(hidden_dim))\n        if fine_observation_matrix:\n            self.observation_matrix = nn.Parameter(\n                0.3 * torch.randn(num_components, hidden_dim, obs_dim)\n            )\n        else:\n            self.observation_matrix = nn.Parameter(\n                0.3 * torch.randn(hidden_dim, obs_dim)\n            )\n        if fine_observation_noise:\n            self.log_obs_noise = nn.Parameter(\n                0.1 * torch.randn(num_components, obs_dim)\n            )\n        else:\n            self.log_obs_noise = nn.Parameter(0.1 * torch.randn(obs_dim))\n\n        # define the prior distribution p(x_0) over the continuous latent at the initial time step t=0\n        x_init_mvn = pyro.distributions.MultivariateNormal(\n            torch.zeros(self.hidden_dim), torch.eye(self.hidden_dim)\n        )\n        self.x_init_mvn = mvn_to_funsor(\n            x_init_mvn,\n            real_inputs=OrderedDict([(\"x_0\", funsor.Reals[self.hidden_dim])]),\n        )\n\n    # we construct the various funsors used to compute the marginal log probability and other model quantities.\n    # these funsors depend on the various model parameters.\n    def get_tensors_and_dists(self):\n        # normalize the transition probabilities\n        trans_logits = self.transition_logits - self.transition_logits.logsumexp(\n            dim=-1, keepdim=True\n        )\n        trans_probs = funsor.Tensor(\n            trans_logits, OrderedDict([(\"s\", funsor.Bint[self.num_components])])\n        )\n\n        trans_mvn = pyro.distributions.MultivariateNormal(\n            torch.zeros(self.hidden_dim), self.log_transition_noise.exp().diag_embed()\n        )\n        obs_mvn = pyro.distributions.MultivariateNormal(\n            torch.zeros(self.obs_dim), self.log_obs_noise.exp().diag_embed()\n        )\n\n        event_dims = (\n            (\"s\",) if self.fine_transition_matrix or self.fine_transition_noise else ()\n        )\n        x_trans_dist = matrix_and_mvn_to_funsor(\n            self.transition_matrix, trans_mvn, event_dims, \"x\", \"y\"\n        )\n        event_dims = (\n            (\"s\",)\n            if self.fine_observation_matrix or self.fine_observation_noise\n            else ()\n        )\n        y_dist = matrix_and_mvn_to_funsor(\n            self.observation_matrix, obs_mvn, event_dims, \"x\", \"y\"\n        )\n\n        return trans_logits, trans_probs, trans_mvn, obs_mvn, x_trans_dist, y_dist\n\n    # compute the marginal log probability of the observed data using a moment-matching approximation\n    @funsor.interpretations.moment_matching\n    def log_prob(self, data):\n        (\n            trans_logits,\n            trans_probs,\n            trans_mvn,\n            obs_mvn,\n            x_trans_dist,\n            y_dist,\n        ) = self.get_tensors_and_dists()\n\n        log_prob = funsor.Number(0.0)\n\n        s_vars = {-1: funsor.Tensor(torch.tensor(0), dtype=self.num_components)}\n        x_vars = {}\n\n        for t, y in enumerate(data):\n            # construct free variables for s_t and x_t\n            s_vars[t] = funsor.Variable(f\"s_{t}\", funsor.Bint[self.num_components])\n            x_vars[t] = funsor.Variable(f\"x_{t}\", funsor.Reals[self.hidden_dim])\n\n            # incorporate the discrete switching dynamics\n            log_prob += dist.Categorical(trans_probs(s=s_vars[t - 1]), value=s_vars[t])\n\n            # incorporate the prior term p(x_t | x_{t-1})\n            if t == 0:\n                log_prob += self.x_init_mvn(value=x_vars[t])\n            else:\n                log_prob += x_trans_dist(s=s_vars[t], x=x_vars[t - 1], y=x_vars[t])\n\n            # do a moment-matching reduction. at this point log_prob depends on (moment_matching_lag + 1)-many\n            # pairs of free variables.\n            if t > self.moment_matching_lag - 1:\n                log_prob = log_prob.reduce(\n                    ops.logaddexp,\n                    {\n                        s_vars[t - self.moment_matching_lag],\n                        x_vars[t - self.moment_matching_lag],\n                    },\n                )\n\n            # incorporate the observation p(y_t | x_t, s_t)\n            log_prob += y_dist(s=s_vars[t], x=x_vars[t], y=y)\n\n        T = data.shape[0]\n        # reduce any remaining free variables\n        for t in range(self.moment_matching_lag):\n            log_prob = log_prob.reduce(\n                ops.logaddexp,\n                {\n                    s_vars[T - self.moment_matching_lag + t],\n                    x_vars[T - self.moment_matching_lag + t],\n                },\n            )\n\n        # assert that we've reduced all the free variables in log_prob\n        assert not log_prob.inputs, \"unexpected free variables remain\"\n\n        # return the PyTorch tensor behind log_prob (which we can directly differentiate)\n        return log_prob.data\n\n    # do filtering, prediction, and smoothing using a moment-matching approximation.\n    # here we implicitly use a moment matching lag of L = 1. the general logic follows\n    # the logic in the log_prob method.\n    @torch.no_grad()\n    @funsor.interpretations.moment_matching\n    def filter_and_predict(self, data, smoothing=False):\n        (\n            trans_logits,\n            trans_probs,\n            trans_mvn,\n            obs_mvn,\n            x_trans_dist,\n            y_dist,\n        ) = self.get_tensors_and_dists()\n\n        log_prob = funsor.Number(0.0)\n\n        s_vars = {-1: funsor.Tensor(torch.tensor(0), dtype=self.num_components)}\n        x_vars = {-1: None}\n\n        predictive_x_dists, predictive_y_dists, filtering_dists = [], [], []\n        test_LLs = []\n\n        for t, y in enumerate(data):\n            s_vars[t] = funsor.Variable(f\"s_{t}\", funsor.Bint[self.num_components])\n            x_vars[t] = funsor.Variable(f\"x_{t}\", funsor.Reals[self.hidden_dim])\n\n            log_prob += dist.Categorical(trans_probs(s=s_vars[t - 1]), value=s_vars[t])\n\n            if t == 0:\n                log_prob += self.x_init_mvn(value=x_vars[t])\n            else:\n                log_prob += x_trans_dist(s=s_vars[t], x=x_vars[t - 1], y=x_vars[t])\n\n            if t > 0:\n                log_prob = log_prob.reduce(\n                    ops.logaddexp, {s_vars[t - 1], x_vars[t - 1]}\n                )\n\n            # do 1-step prediction and compute test LL\n            if t > 0:\n                predictive_x_dists.append(log_prob)\n                _log_prob = log_prob - log_prob.reduce(ops.logaddexp)\n                predictive_y_dist = y_dist(s=s_vars[t], x=x_vars[t]) + _log_prob\n                test_LLs.append(\n                    predictive_y_dist(y=y).reduce(ops.logaddexp).data.item()\n                )\n                predictive_y_dist = predictive_y_dist.reduce(\n                    ops.logaddexp, {f\"x_{t}\", f\"s_{t}\"}\n                )\n                predictive_y_dists.append(funsor_to_mvn(predictive_y_dist, 0, ()))\n\n            log_prob += y_dist(s=s_vars[t], x=x_vars[t], y=y)\n\n            # save filtering dists for forward-backward smoothing\n            if smoothing:\n                filtering_dists.append(log_prob)\n\n        # do the backward recursion using previously computed ingredients\n        if smoothing:\n            # seed the backward recursion with the filtering distribution at t=T\n            smoothing_dists = [filtering_dists[-1]]\n            T = data.size(0)\n\n            s_vars = {\n                t: funsor.Variable(f\"s_{t}\", funsor.Bint[self.num_components])\n                for t in range(T)\n            }\n            x_vars = {\n                t: funsor.Variable(f\"x_{t}\", funsor.Reals[self.hidden_dim])\n                for t in range(T)\n            }\n\n            # do the backward recursion.\n            # let p[t|t-1] be the predictive distribution at time step t.\n            # let p[t|t] be the filtering distribution at time step t.\n            # let f[t] denote the prior (transition) density at time step t.\n            # then the smoothing distribution p[t|T] at time step t is\n            # given by the following recursion.\n            # p[t-1|T] = p[t-1|t-1] <p[t|T] f[t] / p[t|t-1]>\n            # where <...> denotes integration of the latent variables at time step t.\n            for t in reversed(range(T - 1)):\n                integral = smoothing_dists[-1] - predictive_x_dists[t]\n                integral += dist.Categorical(\n                    trans_probs(s=s_vars[t]), value=s_vars[t + 1]\n                )\n                integral += x_trans_dist(s=s_vars[t], x=x_vars[t], y=x_vars[t + 1])\n                integral = integral.reduce(\n                    ops.logaddexp, {s_vars[t + 1], x_vars[t + 1]}\n                )\n                smoothing_dists.append(filtering_dists[t] + integral)\n\n        # compute predictive test MSE and predictive variances\n        predictive_means = torch.stack([d.mean for d in predictive_y_dists])  # T-1 ydim\n        predictive_vars = torch.stack(\n            [d.covariance_matrix.diagonal(dim1=-1, dim2=-2) for d in predictive_y_dists]\n        )\n        predictive_mse = (predictive_means - data[1:, :]).pow(2.0).mean(-1)\n\n        if smoothing:\n            # compute smoothed mean function\n            smoothing_dists = [\n                funsor_to_cat_and_mvn(d, 0, (f\"s_{t}\",))\n                for t, d in enumerate(reversed(smoothing_dists))\n            ]\n            means = torch.stack([d[1].mean for d in smoothing_dists])  # T 2 xdim\n            means = torch.matmul(means.unsqueeze(-2), self.observation_matrix).squeeze(\n                -2\n            )  # T 2 ydim\n\n            probs = torch.stack([d[0].logits for d in smoothing_dists]).exp()\n            probs = probs / probs.sum(-1, keepdim=True)  # T 2\n\n            smoothing_means = (probs.unsqueeze(-1) * means).sum(-2)  # T ydim\n            smoothing_probs = probs[:, 1]\n\n            return (\n                predictive_mse,\n                torch.tensor(np.array(test_LLs)),\n                predictive_means,\n                predictive_vars,\n                smoothing_means,\n                smoothing_probs,\n            )\n        else:\n            return predictive_mse, torch.tensor(np.array(test_LLs))\n\n\ndef main(args):\n    funsor.set_backend(\"torch\")\n\n    # download and pre-process EEG data if not in test mode\n    if not args.test:\n        download_data()\n        N_val, N_test = 149, 200\n        data = np.loadtxt(\"eeg.dat\", delimiter=\",\", skiprows=19)\n        print(f\"[raw data shape] {data.shape}\")\n        data = data[::20, :]\n        print(f\"[data shape after thinning] {data.shape}\")\n        eye_state = [int(d) for d in data[:, -1].tolist()]\n        data = torch.tensor(data[:, :-1]).float()\n    # in test mode (for continuous integration on github) so create fake data\n    else:\n        data = torch.randn(10, 3)\n        N_val, N_test = 2, 2\n\n    T, obs_dim = data.shape\n    N_train = T - N_test - N_val\n\n    np.random.seed(0)\n    rand_perm = np.random.permutation(N_val + N_test)\n    val_indices = rand_perm[0:N_val]\n    test_indices = rand_perm[N_val:]\n\n    data_mean = data[0:N_train, :].mean(0)\n    data -= data_mean\n    data_std = data[0:N_train, :].std(0)\n    data /= data_std\n\n    print(f\"Length of time series T: {T}   Observation dimension: {obs_dim}\")\n    print(f\"N_train: {N_train}  N_val: {N_val}  N_test: {N_test}\")\n\n    torch.manual_seed(args.seed)\n\n    # set up model\n    slds = SLDS(\n        num_components=args.num_components,\n        hidden_dim=args.hidden_dim,\n        obs_dim=obs_dim,\n        fine_observation_noise=args.fon,\n        fine_transition_noise=args.ftn,\n        fine_observation_matrix=args.fom,\n        fine_transition_matrix=args.ftm,\n        moment_matching_lag=args.moment_matching_lag,\n    )\n\n    # set up optimizer\n    adam = torch.optim.Adam(\n        slds.parameters(),\n        lr=args.learning_rate,\n        betas=(args.beta1, 0.999),\n        amsgrad=True,\n    )\n    scheduler = torch.optim.lr_scheduler.ExponentialLR(adam, gamma=args.gamma)\n    ts = [time.time()]\n\n    report_frequency = 1\n\n    # training loop\n    for step in range(args.num_steps):\n        nll = -slds.log_prob(data[0:N_train, :]) / N_train\n        nll.backward()\n\n        if step == 5:\n            scheduler.base_lrs[0] *= 0.20\n\n        adam.step()\n        scheduler.step()\n        adam.zero_grad()\n\n        if step % report_frequency == 0 or step == args.num_steps - 1:\n            step_dt = ts[-1] - ts[-2] if step > 0 else 0.0\n            pred_mse, pred_LLs = slds.filter_and_predict(\n                data[0 : N_train + N_val + N_test, :]\n            )\n            val_mse = pred_mse[val_indices].mean().item()\n            test_mse = pred_mse[test_indices].mean().item()\n            val_ll = pred_LLs[val_indices].mean().item()\n            test_ll = pred_LLs[test_indices].mean().item()\n\n            stats = \"[step %03d] train_nll: %.5f val_mse: %.5f val_ll: %.5f test_mse: %.5f test_ll: %.5f\\t(dt: %.2f)\"\n            print(\n                stats % (step, nll.item(), val_mse, val_ll, test_mse, test_ll, step_dt)\n            )\n\n        ts.append(time.time())\n\n    # plot predictions and smoothed means\n    if args.plot:\n        assert not args.test\n        (\n            predicted_mse,\n            LLs,\n            pred_means,\n            pred_vars,\n            smooth_means,\n            smooth_probs,\n        ) = slds.filter_and_predict(data, smoothing=True)\n\n        pred_means = pred_means.data.numpy()\n        pred_stds = pred_vars.sqrt().data.numpy()\n        smooth_means = smooth_means.data.numpy()\n        smooth_probs = smooth_probs.data.numpy()\n\n        import matplotlib\n\n        matplotlib.use(\"Agg\")  # noqa: E402\n        import matplotlib.pyplot as plt\n\n        f, axes = plt.subplots(4, 1, figsize=(12, 8), sharex=True)\n        T = data.size(0)\n        N_valtest = N_val + N_test\n        to_seconds = 117.0 / T\n\n        for k, ax in enumerate(axes[:-1]):\n            which = [0, 4, 10][k]\n            ax.plot(to_seconds * np.arange(T), data[:, which], \"ko\", markersize=2)\n            ax.plot(\n                to_seconds * np.arange(N_train),\n                smooth_means[:N_train, which],\n                ls=\"solid\",\n                color=\"r\",\n            )\n\n            ax.plot(\n                to_seconds * (N_train + np.arange(N_valtest)),\n                pred_means[-N_valtest:, which],\n                ls=\"solid\",\n                color=\"b\",\n            )\n            ax.fill_between(\n                to_seconds * (N_train + np.arange(N_valtest)),\n                pred_means[-N_valtest:, which] - 1.645 * pred_stds[-N_valtest:, which],\n                pred_means[-N_valtest:, which] + 1.645 * pred_stds[-N_valtest:, which],\n                color=\"lightblue\",\n            )\n            ax.set_ylabel(f\"$y_{which + 1}$\", fontsize=20)\n            ax.tick_params(axis=\"both\", which=\"major\", labelsize=14)\n\n        axes[-1].plot(to_seconds * np.arange(T), eye_state, \"k\", ls=\"solid\")\n        axes[-1].plot(to_seconds * np.arange(T), smooth_probs, \"r\", ls=\"solid\")\n        axes[-1].set_xlabel(\"Time (s)\", fontsize=20)\n        axes[-1].set_ylabel(\"Eye state\", fontsize=20)\n        axes[-1].tick_params(axis=\"both\", which=\"major\", labelsize=14)\n\n        plt.tight_layout(pad=0.7)\n        plt.savefig(\"eeg.pdf\")\n\n\nif __name__ == \"__main__\":\n    parser = argparse.ArgumentParser(description=\"Switching linear dynamical system\")\n    parser.add_argument(\"-n\", \"--num-steps\", default=3, type=int)\n    parser.add_argument(\"-s\", \"--seed\", default=15, type=int)\n    parser.add_argument(\"-hd\", \"--hidden-dim\", default=5, type=int)\n    parser.add_argument(\"-k\", \"--num-components\", default=2, type=int)\n    parser.add_argument(\"-lr\", \"--learning-rate\", default=0.5, type=float)\n    parser.add_argument(\"-b1\", \"--beta1\", default=0.75, type=float)\n    parser.add_argument(\"-g\", \"--gamma\", default=0.99, type=float)\n    parser.add_argument(\"-mml\", \"--moment-matching-lag\", default=1, type=int)\n    parser.add_argument(\"--plot\", action=\"store_true\")\n    parser.add_argument(\"--fon\", action=\"store_true\")\n    parser.add_argument(\"--ftm\", action=\"store_true\")\n    parser.add_argument(\"--fom\", action=\"store_true\")\n    parser.add_argument(\"--ftn\", action=\"store_true\")\n    parser.add_argument(\"--test\", action=\"store_true\")\n    args = parser.parse_args()\n\n    main(args)"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.7.9"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}